unit tiltefeitos;

//=== Unit para utilizao de efeitos em imagens bmp
//=== http://www.tilt.net

interface

uses
  windows, contnrs, classes, graphics, jpeg, extCtrls, sysUtils, math;

const
  CONVOKERNELX = 3;
  CONVOKERNELY = 3;

  CONVOMEIOX = CONVOKERNELX div 2;
  CONVOMEIOY = CONVOKERNELY div 2;

type
  TConvoInfo = record
        r: single;
        g: single;
        b: single;
  end;
  TConvoMatrix = array[-CONVOMEIOX..CONVOMEIOX, -CONVOMEIOY..CONVOMEIOY] of TConvoInfo;

  TConvoIndexMatrix = array[-CONVOMEIOX..CONVOMEIOX, -CONVOMEIOY..CONVOMEIOY] of single;

  procedure setConvoMatrix(var mat: TConvoMatrix; r, g, b: single);
  procedure fxFiltro(bmp: TBitmap; cor: TColor);
  procedure fxBrilho(bmp: TBitmap; valor: byte);
  procedure fxEspelharHorz(bmp: TBitmap);
  procedure fxEspelharVert(bmp: TBitmap);
  procedure fxScanline(bmp: TBitmap);
  procedure fxBlur(bmp: TBitmap);
  procedure fxConvolution(bmp: TBitmap; matrix: TConvoMatrix; divisor, bias: single);
  procedure fxEmboss(Bmp: TBitmap);
  procedure fxSharpen(bmp: TBitmap);
  procedure fxCopiarComAlfa(src, dst: TBitmap; x, y: integer; alfa: byte);
  procedure fxGrayscale(bmp: TBitmap);

implementation

function RGBToColor(r, g, b: byte): TColor;
begin
  result:= (b shl 16) or (g shl 8) or (r);
end;

procedure colortorgb(c: tcolor; var r, g, b: byte);
begin
  r:= byte(c);
  g:= byte(c shr 8);
  b:= byte(c shr 16);
end;

procedure setScanline(bmp: TBitmap; x, y: integer; val: TColor);
var
  lin: PByteArray;
  r,g,b: byte;
begin
  lin:= bmp.ScanLine[y];

  colortorgb(val, b, g, r);

  lin[x*3]:= r;
  lin[x*3+1]:= g;
  lin[x*3+2]:= b;
end;

function getScanline(bmp: TBitmap; x, y: integer): TColor;
var
  lin: PByteArray;
  r,g,b: byte;
begin
  lin:= bmp.ScanLine[y];
  r:= lin[x*3];
  g:= lin[x*3+1];
  b:= lin[x*3+2];
  result:= RGBToColor(b, g, r);
end;

procedure setConvoItem(var mat: TConvoMatrix; x, y: integer; val: single);
begin
  mat[x, y].r:= val;
  mat[x, y].g:= val;
  mat[x, y].b:= val;
end;

procedure setConvoMatrix(var mat: TConvoMatrix; r, g, b: single);
var
  x,y: integer;
begin
  for x:= -CONVOMEIOX to CONVOMEIOX do
      for y:= -CONVOMEIOY to CONVOMEIOY do begin
          mat[x, y].r:= r;
          mat[x, y].g:= g;
          mat[x, y].b:= b;
      end;
end;

procedure indexMatToConvo(var convo: TConvoMatrix; src: TConvoIndexMatrix);
var
  x,y: integer;
begin
  for x:= -CONVOMEIOX to CONVOMEIOX do
      for y:= -CONVOMEIOY to CONVOMEIOY do begin
          convo[x, y].r:= src[x, y];
          convo[x, y].g:= src[x, y];
          convo[x, y].b:= src[x, y];
      end;
end;

function getConvoMatrix(src: TConvoIndexMatrix): TConvoMatrix;
var
  x,y: integer;
begin
  for x:= -CONVOMEIOX to CONVOMEIOX do
      for y:= -CONVOMEIOY to CONVOMEIOY do begin
          result[x, y].r:= src[x, y];
          result[x, y].g:= src[x, y];
          result[x, y].b:= src[x, y];
      end;
end;

//== Adiciona brilho  uma imagem.
//== Valor corresponde  "potncia" do brilho.
procedure fxBrilho(bmp: TBitmap; valor: byte);
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;
  fxFiltro(bmp, RGBToColor(valor, valor, valor));
end;

//== Adiciona um "filtro" de cor  uma imagem.
procedure fxFiltro(bmp: TBitmap; cor: TColor);
var
  r,g,b: byte;
  ir,ig,ib: single;
  x,y: integer;
  nr,ng,nb: single;
  c: TColor;
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;
  colortorgb(cor, b, g, r);
  ir:= r / 255;
  ig:= g / 255;
  ib:= b / 255;
  for x:= 0 to bmp.width -1 do
      for y:= 0 to bmp.height -1 do begin
          c:= getScanline(bmp, x, y);
          colortorgb(c, b, g, r);
          nr:= r + (r * ir); if nr > 255 then nr:= 255;
          ng:= g + (g * ig); if ng > 255 then ng:= 255;
          nb:= b + (b * ib); if nb > 255 then nb:= 255;
          c:= RGBToColor(trunc(nb), trunc(ng), trunc(nr));
          setScanline(bmp, x, y, c);
      end;
end;

//== Espelha uma imagem horizontalmente (troca as colunas)
procedure fxEspelharHorz(bmp: TBitmap);
var
  x,y: integer;
  aux: TColor;
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;
  for y:= 0 to bmp.Height -1 do
      for x:= 0 to (bmp.Width -1) div 2 do begin
          aux:= getScanline(bmp, bmp.width -1 -x, y);
          setScanline(bmp, bmp.width -1 -x, y, getScanline(bmp, x, y));
          setScanline(bmp, x, y, aux);
      end;
end;

procedure fxEspelharVert(bmp: TBitmap);
var
  aux: TColor;
  x,y: integer;
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;
  for x:= 0 to bmp.width -1 do
  for y:= 0 to (bmp.height -1) div 2 do begin
    aux:= getScanline(bmp, x, y);
    setScanline(bmp,x,y,getScanline(bmp,x,bmp.height -1 -y));
    setScanline(bmp, x, bmp.height -1 -y, aux);
  end;
end;

procedure fxScanline(bmp: TBitmap);
var
  x,y: integer;
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;
  y:= 0;
  while (y < bmp.height) do begin
    for x:= 0 to bmp.width -1 do
        setScanline(bmp, x, y, clBlack);
    y:= y + 2;
  end;
end;

procedure fxConvolution(bmp: TBitmap; matrix: TConvoMatrix; divisor, bias: single);
var
  buffer: TBitmap;
  somar, somab, somag: extended;
  r,g,b: byte;
  x,y: integer;
  ax,ay: integer;
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;

  buffer:= TBitmap.create;
  buffer.Assign(bmp);
  for y:= 0 to bmp.height -1 do begin
      for x:= 0 to bmp.Width -1 do begin
          somar:= 0;
          somag:= 0;
          somab:= 0;
          //pegar a amostra
          for ax:= -CONVOMEIOX to CONVOMEIOX do
              for ay:= -CONVOMEIOY to CONVOMEIOY do begin
                  if ((x + ax) < 0) or ((x + ax) > bmp.width -1) then continue;
                  if ((y + ay) < 0) or ((y + ay) > bmp.height -1) then continue;
                  colortorgb(getScanline(buffer, x + ax, y + ay), r, g, b);
                  somar:= somar + r * matrix[ax][ay].r;
                  somag:= somag + g * matrix[ax][ay].g;
                  somab:= somab + b * matrix[ax][ay].b;
              end;

          somar:= somar / divisor + bias;
          somag:= somag / divisor + bias;
          somab:= somab / divisor + bias;
          if (somar < 0) then somar:= 0;
          if (somag < 0) then somag:= 0;
          if (somab < 0) then somab:= 0;
          if (somar > 255) then somar:= 255;
          if (somag > 255) then somag:= 255;
          if (somab > 255) then somab:= 255;
          r:= trunc((somar));
          g:= trunc((somag));
          b:= trunc((somab));
          setScanline(bmp, x, y, RGBToColor(r, g, b));
          end;
  end;
  buffer.free;
end;

//== Executa um blur do tipo mdia (average).
procedure fxBlur(bmp: TBitmap);
var
  mat: TConvoIndexMatrix;
begin
  mat[-1,-1]:=  1; mat[ 0,-1]:=  1; mat[+1,-1]:=  1;
  mat[-1, 0]:=  1; mat[ 0, 0]:=  1; mat[+1, 0]:=  1;
  mat[-1,+1]:=  1; mat[ 0,+1]:=  1; mat[+1,+1]:=  1;
  fxConvolution(bmp, getConvoMatrix(mat), 9, 0 );
end;

//== Embossa a imagem
procedure fxEmboss(Bmp: TBitmap);
var
  mat: TConvoIndexMatrix;
begin
  mat[-1,-1]:= -1; mat[ 0,-1]:= -1; mat[+1,-1]:=  0;
  mat[-1, 0]:= -1; mat[ 0, 0]:=  0; mat[+1, 0]:= +1;
  mat[-1,+1]:=  0; mat[ 0,+1]:= +1; mat[+1,+1]:= +1;
  fxConvolution(bmp, getConvoMatrix(mat), 1, 128 );
end;

//== Sharpen (o contrrio de blur)
procedure fxSharpen(bmp: TBitmap);
var
  mat: TConvoIndexMatrix;
begin
  mat[-1,-1]:= -1; mat[ 0,-1]:=  0; mat[+1,-1]:=  0;
  mat[-1, 0]:=  0; mat[ 0, 0]:=  3; mat[+1, 0]:=  0;
  mat[-1,+1]:=  0; mat[ 0,+1]:=  0; mat[+1,+1]:= -1;
  fxConvolution(bmp, getConvoMatrix(mat), 1, 0 );
end;

//== Source-alfa (src  copiado para dst(x,y) com alfa
procedure fxCopiarComAlfa(src, dst: TBitmap;
          x, y: integer; alfa: byte);
var
  cx,cy,ax,ay: integer;
  sr,sg,sb,dr,dg,db: byte;
begin
  if src.PixelFormat <> pf24bit then
     src.PixelFormat:= pf24bit;
  if dst.PixelFormat <> pf24bit then
     dst.PixelFormat:= pf24bit;
  if x >= dst.Width then exit;
  if y >= dst.height then exit;
  ay:= 0;
  for cy:= y to min(dst.height-1,y+src.height-1) do begin
    ax:= 0;
    for cx:= x to min(dst.width-1,x+src.width -1) do begin
      colortorgb(getScanline(src, ax, ay), sr, sg, sb);
      colortorgb(getScanline(dst, cx, cy), dr, dg, db);
      dr:= dr + ((sr - dr) * alfa) shr 8;
      dg:= dg + ((sg - dg) * alfa) shr 8;
      db:= db + ((sb - db) * alfa) shr 8;
      setScanline(dst, cx, cy, RGBToColor(dr, dg, db));
      ax:= ax + 1;
    end;
    ay:= ay + 1;
  end;
end;

procedure fxGrayscale(bmp: TBitmap);
var
  x,y: integer;
  r,g,b,nr,ng,nb: byte;
begin
  if bmp.PixelFormat <> pf24bit then
     bmp.PixelFormat:= pf24bit;
  for x:= 0 to bmp.width -1 do
  for y:= 0 to bmp.height -1 do begin
    colortorgb(getScanline(bmp, x, y), r, g, b);
    nr:= trunc(r * 0.2125 + g * 0.7154 + b * 0.0721);
    ng:= nr;
    nb:= nr;
    setScanline(bmp, x, y, rgbToColor(nr, ng, nb));
  end;
end;

end.
